from typing import Tuple
from easyansi.core import core as _core
from easyansi.common import field_validations as _validator
import shutil

# screen size
DEFAULT_COLS = 80
DEFAULT_ROWS = 24

# clear screen
CLEAR_SCREEN = f"{_core.CSI}2J"
CLEAR = CLEAR_SCREEN
CLEAR_SCREEN_FWD = f"{_core.CSI}0J"
CLEAR_FWD = CLEAR_SCREEN_FWD
CLEAR_SCREEN_BWD = f"{_core.CSI}1J"
CLEAR_BWD = CLEAR_SCREEN_BWD

# clear row
CLEAR_ROW = f"{_core.CSI}2K"
CLEAR_ROW_FWD = f"{_core.CSI}0K"
CLEAR_ROW_BWD = f"{_core.CSI}1K"

# reset
RESET = _core.RESET


def size() -> Tuple[int, int]:
    """Return the screen size in (cols, rows).

    This is not an ANSI function, but uses python to retrieve this for you."""
    screen_size = shutil.get_terminal_size()
    return screen_size.columns, screen_size.lines


def sufficient_size(min_cols: int, min_rows: int) -> Tuple[bool, int, int]:
    """Given a minimum number of columns and rows, check that the screen is at least this size.
    Returns True / False if the size meets the minimums, followed by the current number of columns and rows.

    Parameters:
        min_cols: The minimum number of columns for the terminal size.
        min_rows: The minimum number of rows for the terminal size.
    """
    minimum_columns = 1
    minimum_rows = 1
    _validator.check_int_minimum_value(min_cols, minimum_columns, "Minimum number of columns")
    _validator.check_int_minimum_value(min_rows, minimum_rows, "Minimum number of rows")
    cols, rows = size()
    if (cols < min_cols) or (rows < min_rows):
        return False, cols, rows
    return True, cols, rows



from typing import Tuple, Optional
from easyansi.core import core as _core
from easyansi.common import field_validations as _validator
import sys
_WINDOWS = False
try:
    import tty
    import termios
except ModuleNotFoundError:
    import msvcrt
    _WINDOWS = True

MIN_MOVEMENT = 1
MIN_COL = 0
MIN_ROW = 0


def up(rows: int = 1) -> str:
    """Given a number of rows, move the cursor up."""
    _validator.check_int_minimum_value(rows, MIN_MOVEMENT, "Cursor up row count")
    return f"{_core.CSI}{rows}A"


def down(rows: int = 1) -> str:
    """Given a number of rows, move the cursor down."""
    _validator.check_int_minimum_value(rows, MIN_MOVEMENT, "Cursor down row count")
    return f"{_core.CSI}{rows}B"


def left(cols: int = 1) -> str:
    """Given a number of columns, move the cursor left."""
    _validator.check_int_minimum_value(cols, MIN_MOVEMENT, "Cursor left column count")
    return f"{_core.CSI}{cols}D"


def right(cols: int = 1) -> str:
    """Given a number of columns, move the cursor right."""
    _validator.check_int_minimum_value(cols, MIN_MOVEMENT, "Cursor right column count")
    return f"{_core.CSI}{cols}C"


def move(cols: Optional[int] = None, rows: Optional[int] = None) -> str:
    """Given a number of columns and/or rows, move the cursor relative to its current position.
    If a value is zero, this is the same as setting it to None.

    Parameters:
        cols: Negative number = left movement, Positive number = right movement
        rows: Negative number = up movement, Positive number = down movement
    """
    _validator.check_if_any_have_value("Cursor move values", cols, rows, zero_int_no_value=True)
    move_code = ""
    if cols is not None:
        if isinstance(cols, int):
            if cols < 0:
                move_code += left(cols * -1)
            elif cols > 0:
                move_code += right(cols)
            # else:
                # if cols == 0, skip left/right code like it was None
        else:
            # let the right() function throw an error
            right(cols)
    if rows is not None:
        if isinstance(rows, int):
            if rows < 0:
                move_code += up(rows * -1)
            elif rows > 0:
                move_code += down(rows)
            # else:
                # if rows == 0, skip up/down code like it was None
        else:
            # let the down() function throw an error
            move_code += down(rows)
    return move_code


def next_line(rows: int = 1) -> str:
    """Given a number of rows, move cursor down and to the first column."""
    _validator.check_int_minimum_value(rows, MIN_MOVEMENT, "Cursor next line row count")
    return f"{_core.CSI}{rows}E"


def previous_line(rows: int = 1) -> str:
    """Given a number of rows, move cursor up and to the first column."""
    _validator.check_int_minimum_value(rows, MIN_MOVEMENT, "Cursor previous line row count")
    return f"{_core.CSI}{rows}F"


def locate(col: int, row: int) -> str:
    """Given a column and row, move the cursor to this location."""
    _validator.check_int_minimum_value(col, MIN_COL, "Cursor locate column value")
    _validator.check_int_minimum_value(row, MIN_ROW, "Cursor locate row value")
    return f"{_core.CSI}{row + 1};{col + 1}H"


def locate_col(col: int) -> str:
    """Given a column, move the cursor to this column on the current row."""
    _validator.check_int_minimum_value(col, MIN_COL, "Cursor locate_col column value")
    return f"{_core.CSI}{col + 1}G"


def position() -> Tuple[int, int]:
    """Return the col, row coordinates of the cursor."""
    response = _get_position_response()
    loc = response.split(";")
    col = int(loc[1]) - 1
    row = int(loc[0]) - 1
    return col, row


def _get_position_response() -> str:
    """Make the actual ANSI call for the cursor position.
    This code is separated out to make it easy to mock in unit tests."""
    from easyansi.core import utils
    response = ""
    utils.prnt(f"{_core.CSI}6n")  # Send ANSI request for cursor location
    if _WINDOWS:
        while msvcrt.kbhit():  # type: ignore
            char_in = msvcrt.getch()  # type: ignore
            response += char_in.decode("utf-8")
        response = response[2:-1]  # We do not need the first 2 bytes or the last byte
    else:
        f = sys.stdin.fileno()
        terminal_settings = termios.tcgetattr(f)  # Save terminal settings
        try:
            tty.setraw(f)
            sys.stdin.read(2)  # We do not need the first 2 bytes
            char_in = sys.stdin.read(1)
            while char_in != "R":  # R will be the last character of the ANSI response, and we don't need it
                response += char_in
                char_in = sys.stdin.read(1)  # Read a single character
        finally:
            termios.tcsetattr(f, termios.TCSADRAIN, terminal_settings)  # Restore terminal settings
    return response


###########################################################################
# Convenience Constants
###########################################################################

UP = up(1)
DOWN = down(1)
LEFT = left(1)
RIGHT = right(1)

NEXT_LINE = next_line(1)
PREVIOUS_LINE = previous_line(1)

HOME = locate(0, 0)
HOME_ON_SCREEN = HOME
HOME_ON_ROW = locate_col(0)

HIDE = f"{_core.CSI}?25l"
HIDE_ON = HIDE
SHOW = f"{_core.CSI}?25h"
SHOW_ON = SHOW
HIDE_OFF = SHOW
SHOW_OFF = HIDE



from enum import Enum
from easyansi.core.core import CSI
from typing import Optional, Union, Any, Tuple, List, cast
from easyansi.common import field_validations as _validator

# Default Colors
DEFAULT_IDX = -1
DEFAULT_IDX_STR = "-1"
DEFAULT_FG = f"{CSI}39m"
DEFAULT_BG = f"{CSI}49m"
DEFAULT = DEFAULT_FG

MIN_VALUE = 0

# Color Types
_COLOR_TYPE_INTEGER = "integer"
_COLOR_TYPE_HEX_STRING = "hex_string"
_COLOR_TYPE_SEQUENCE = "sequence"
_COLOR_TYPE_UNKNOWN = "unknown"


class ColorBuilder(Enum):

    PALETTE_16 = (16, 15, 1, (1, -1))
    PALETTE_256 = (256, 255, 2, (1, -1))
    PALETTE_RGB = (16777216, 255, 6, (1, 3))

    def __init__(self, palette: int, max_color_value: int, max_hex_digits: int,
                 sequence_len: Tuple[int, int]):
        self._palette = palette
        self._max_color_seq = palette - 1
        self._max_color_value = max_color_value
        self._max_hex_digits = max_hex_digits
        self._sequence_len = sequence_len

    @property
    def min_value(self) -> int:
        """Return the minimum index value for a color palette."""
        return MIN_VALUE

    @property
    def max_value(self) -> int:
        """Return the maximum index value for a color palette."""
        return self._max_color_value

    @property
    def max_value_seq(self) -> int:
        """Return the maximum integer value for a color palette within a sequence."""
        return self._max_color_seq

    @property
    def default_value(self) -> int:
        """Return the default index value for a color palette palette."""
        return DEFAULT_IDX

    @property
    def palette(self) -> int:
        """Return the total number of colors for the palette."""
        return self._palette

    ######################################################################################
    # Main color entry point
    ######################################################################################

    def color(self,
              fg: Optional[Union[int, str, List[Union[int, str]], Tuple[Union[int, str]]]] = None,
              bg: Optional[Union[int, str, List[Union[int, str]], Tuple[Union[int, str]]]] = None) -> str:
        """Return the ANSI foreground and/or background color code(s).

        Parameters:
            fg: The foreground color value in one of several supported formats.
            bg: The background color value in one of several supported formats.
        """
        _validator.check_if_any_have_value("Color Values", fg, bg)
        ansi_code = ""
        if fg is not None:
            ansi_code += self._get_color_code(fg, False)
        if bg is not None:
            ansi_code += self._get_color_code(bg, True)
        return ansi_code

    def _get_color_code(self, color_value: Union[int, str, List[Union[int, str]], Tuple[Union[int, str]]],
                        is_bg: bool) -> str:
        """Return the ANSI color code for the value provided.

        Parameters:
            color_value: The color value in one of several supported formats.
            is_bg: True = Return background color code, else return foreground color code.
        """
        color_code = None
        color_type = self._get_color_value_type(color_value)
        if color_type == _COLOR_TYPE_INTEGER:
            color_value = cast(int, color_value)
            self._validate_int_color_value(color_value, is_bg)
            color_code = self._get_color_code_by_int(color_value, is_bg)
        elif color_type == _COLOR_TYPE_HEX_STRING:
            color_value = cast(str, color_value)
            self._validate_hex_color_value(color_value, is_bg)
            color_code = self._get_color_code_by_hex(color_value, is_bg)
        elif color_type == _COLOR_TYPE_SEQUENCE:
            color_value = cast(Union[List[Union[int, str]], Tuple[Union[int, str]]], color_value)
            self._validate_sequence_color_value(color_value, is_bg)
            color_code = self._get_color_code_by_sequence(color_value, is_bg)
        else:  # color_type == _COLOR_TYPE_UNKNOWN
            self._unknown_color_value_error(color_value, is_bg)
        color_code = cast(str, color_code)
        return color_code

    @staticmethod
    def _get_color_value_type(color_value: Union[int, str, List[Union[int, str]], Tuple[Union[int, str]]]) -> str:
        """Given a color value, determine the type of the data.

        Parameters:
            color_value: The value to determine the type of data.
        """
        if (isinstance(color_value, int)) and (not isinstance(color_value, bool)):
            return _COLOR_TYPE_INTEGER
        elif isinstance(color_value, str):
            return _COLOR_TYPE_HEX_STRING
        elif (isinstance(color_value, list)) or (isinstance(color_value, tuple)):
            return _COLOR_TYPE_SEQUENCE
        else:
            return _COLOR_TYPE_UNKNOWN

    @staticmethod
    def _unknown_color_value_error(color_value: Any, is_bg: bool) -> None:
        """Could not determine the type of the color value, throw an error.

        Parameters:
            color_value: The color value in question.
            is_bg: True = this is a background color, False = this is a foreground color.
        """
        field = "Background Color Type" if is_bg else "Foreground Color Type"
        msg = "Unknown color value type"
        _validator._raise_error(field, str(color_value), msg, TypeError)

    ######################################################################################
    # Process integer color values
    ######################################################################################

    def _validate_int_color_value(self, color_value: int, is_bg: bool) -> None:
        """Given a color value, determine if it is a valid integer value.

        Parameters:
            color_value: The color value to check to be a valid color integer.
            is_bg: True = this is a background color, False = this is a foreground color.
        """
        field = "Color Background Integer Value" if is_bg else "Color Foreground Integer Value"
        _validator.check_int_in_range(color_value, DEFAULT_IDX, self.max_value_seq, field)

    def _get_color_code_by_int(self, color_value: int, is_bg: bool) -> str:
        """Return the ANSI color code for the integer value provided.

        Parameters:
            color_value: The integer color value.
            is_bg: True = Return background color code, False = return foreground color code.
        """
        if color_value == DEFAULT_IDX:
            return DEFAULT_BG if is_bg else DEFAULT_FG
        elif self.palette == ColorBuilder.PALETTE_16.palette:
            color_code = 30 + (is_bg * 10) + ((color_value // 8) * 60) + (color_value % 8)
            return f"{CSI}{color_code}m"
        elif self.palette == ColorBuilder.PALETTE_256.palette:
            ground_code = "48" if is_bg else "38"
            return f"{CSI}{ground_code};5;{color_value}m"
        else:  # RGB palette
            digits = []
            temp_value = color_value
            for i in range(0, 6):
                power = 5 - i
                digits.append(temp_value // (16 ** power))
                temp_value -= digits[i] * (16 ** power)
            red = (digits[0] * 16) + digits[1]
            green = (digits[2] * 16) + digits[3]
            blue = (digits[4] * 16) + digits[5]
            return self._get_rgb_color_code_by_digits(red, green, blue, is_bg)

    @staticmethod
    def _get_rgb_color_code_by_digits(red: int, green: int, blue: int, is_bg: bool) -> str:
        """Given rgb values, return the RGB ANSI color code.

        Parameters:
             red: The red value (0 - 255)
             green: The green value (0 - 255)
             blue: The blue value (0 - 255)
             is_bg: True = Return background color code, False = return foreground color code.
        """
        ground_code = "48" if is_bg else "38"
        return f"{CSI}{ground_code};2;{red};{green};{blue}m"

    ######################################################################################
    # Process string hex color values
    ######################################################################################

    def _validate_hex_color_value(self, color_value: str, is_bg: bool) -> None:
        """Given a string color value, validate it is in a proper hex format.

        Parameters:
            color_value: The color value to check.
            is_bg: True = this is a background color, False = this is a foreground color
        """
        if not (isinstance(color_value, str) and color_value == DEFAULT_IDX_STR):
            field = "Color Background Hex String Value" if is_bg else "Color Foreground Hex String Value"
            hex_regex_1 = r"^[ ]*#?[ ]*[0-9a-fA-F]{1," + str(self._max_hex_digits) + r"}[ ]*$"
            hex_regex_2 = r"^[ ]*0x[0-9a-fA-F]{1," + str(self._max_hex_digits) + r"}[ ]*$"
            try:
                _validator.check_if_string_matches_regex(color_value, hex_regex_1, field)
            except ValueError:
                _validator.check_if_string_matches_regex(color_value, hex_regex_2, field)

    def _get_color_code_by_hex(self, color_value: str, is_bg: bool) -> str:
        """Return the ANSI color code for the string hex code value provided.

        Parameters:
            color_value: The string hex code color value.
            is_bg: True = Return background color code, False = return foreground color code.
        """
        if color_value == DEFAULT_IDX_STR:
            return self._get_color_code_by_int(DEFAULT_IDX, is_bg)
        int_value = self._hex_to_int(color_value)
        return self._get_color_code_by_int(int_value, is_bg)

    @staticmethod
    def _hex_to_int(hex_value: str) -> int:
        """Given a hex string code, return the integer value.

        Parameters:
            hex_value: The string hex code color value.
        """
        hex_vl = hex_value.replace("0x", "")
        hex_vl = hex_vl.replace("#", "")
        hex_vl = hex_vl.replace(" ", "")
        return int(hex_vl, 16)

    ######################################################################################
    # Process color sequences
    ######################################################################################

    def _validate_sequence_color_value(self, color_value: Union[List[Union[int, str]], Tuple[Union[int, str]]],
                                       is_bg: bool) -> None:
        """Given a color_value sequence, validate it and its values.

        Parameters:
            color_value: The color sequence to check
            is_bg: True = this is a background color, False = this is a foreground color
        """
        field = "Color Background Sequence Length" if is_bg else "Color Foreground Sequence Length"
        color_value_len = len(color_value)
        if self.palette == ColorBuilder.PALETTE_RGB.palette:
            try:
                _validator.check_int_in_range(color_value_len, self._sequence_len[0], self._sequence_len[0], field)
            except (ValueError, TypeError):
                _validator.check_int_in_range(color_value_len,
                                              self._sequence_len[1],
                                              self._sequence_len[1], field)
        else:
            _validator.check_int_in_range(color_value_len, self._sequence_len[0], self._sequence_len[0], field)
        for color_vl in color_value:
            color_type = self._get_color_value_type(color_vl)
            if color_type == _COLOR_TYPE_INTEGER:
                if (self.palette == ColorBuilder.PALETTE_RGB.palette and
                        color_value_len == self._sequence_len[1]):
                    color_vl = cast(int, color_vl)
                    ColorBuilder.PALETTE_256._validate_int_color_value(color_vl, is_bg)
                else:
                    color_vl = cast(int, color_vl)
                    self._validate_int_color_value(color_vl, is_bg)
            elif color_type == _COLOR_TYPE_HEX_STRING:
                if (self.palette == ColorBuilder.PALETTE_RGB.palette and
                        color_value_len == self._sequence_len[1]):
                    color_vl = cast(str, color_vl)
                    ColorBuilder.PALETTE_256._validate_hex_color_value(color_vl, is_bg)
                else:
                    color_vl = cast(str, color_vl)
                    self._validate_hex_color_value(color_vl, is_bg)
            else:
                self._unknown_color_value_error(color_vl, is_bg)

    def _get_color_code_by_sequence(self, color_value: Union[List[Union[int, str]], Tuple[Union[int, str]]],
                                    is_bg: bool) -> str:
        """Return the ANSI color code for the list/tuple value provided.
        The sequence should contain either integer or hex strings.

        Parameters:
            color_value: The sequence color value(s).
            is_bg: True = Return background color code, False = return foreground color code.
        """
        if (DEFAULT_IDX in color_value) or (DEFAULT_IDX_STR in color_value):
            return DEFAULT_BG if is_bg else DEFAULT_FG
        color_value_len = len(color_value)
        if color_value_len == self._sequence_len[0]:
            color_vl = color_value[0]
            color_type = self._get_color_value_type(color_vl)
            if color_type == _COLOR_TYPE_INTEGER:
                color_vl = cast(int, color_vl)
                return self._get_color_code_by_int(color_vl, is_bg)
            else:  # color type == string
                color_vl = cast(str, color_vl)
                return self._get_color_code_by_hex(color_vl, is_bg)
        else:  # sequence length = 3
            digits = []
            for color_vl in color_value:
                color_type = self._get_color_value_type(color_vl)
                if color_type == _COLOR_TYPE_INTEGER:
                    color_vl = cast(int, color_vl)
                    digits.append(color_vl)
                else:  # color type = string
                    color_vl = cast(str, color_vl)
                    digits.append(self._hex_to_int(color_vl))
            return self._get_rgb_color_code_by_digits(digits[0], digits[1], digits[2], is_bg)



from enum import Enum
from easyansi.drawing.line_chars import LineChars
from easyansi.drawing.lines import Lines
from easyansi.common import field_validations as _validator
from easyansi.core import cursor as cur

MIN_WIDTH = 2
MIN_HEIGHT = 2


class Shapes(Enum):

    ASCII = (LineChars.ASCII, Lines.ASCII)
    BLOCK = (LineChars.BLOCK, Lines.BLOCK)
    SINGLE = (LineChars.SINGLE, Lines.SINGLE)
    DOUBLE = (LineChars.DOUBLE, Lines.DOUBLE)

    def __init__(self, char_set: LineChars, line_set: Lines):
        self._char_set = char_set
        self._line_set = line_set

    def rectangle(self, width: int, height: int) -> str:
        """Given a width and height, draw a rectangle.

        Parameters:
            width: The width of the box to draw.
            height: The height of the box to draw.
            """
        _validator.check_int_minimum_value(width, MIN_WIDTH, "Rectangle Width")
        _validator.check_int_minimum_value(height, MIN_HEIGHT, "Rectangle Height")
        between_top_corners = width - 2
        between_side_corners = height - 2
        # Top left corner
        rect = self._char_set.top_left
        # Top side
        if between_top_corners > 0:
            rect += self._line_set.horizontal(between_top_corners)
        # Top right corner
        rect += self._char_set.top_right
        rect += cur.move(-width, 1)
        if between_side_corners > 0:
            # Left side
            rect += self._line_set.vertical(between_side_corners)
            rect += cur.move(between_top_corners, -(between_side_corners - 1))
            # Right side
            rect += self._line_set.vertical(between_side_corners)
            rect += cur.move(-width, 1)
        # Bottom left corner
        rect += self._char_set.bottom_left
        # Bottom side
        if between_top_corners > 0:
            rect += self._line_set.horizontal(between_top_corners)
        # Bottom right corner
        rect += self._char_set.bottom_right
        return rect

    def square(self, size: int) -> str:
        """Given a size, draw a square.
        This is a convenience method that calls rectangle with the same width and height.

        Parameters:
            size: The width and height of the square.
        """
        return self.rectangle(size, size)



from enum import Enum
from typing import Dict

LINE_CHARS_KEYS = ("top_left", "top_right", "bottom_left", "bottom_right",
                   "center_cross", "left_cross", "right_cross", "top_cross", "bottom_cross",
                   "horizontal", "vertical",
                   "diagonal_fwd", "diagonal_bwd", "diagonal_cross")


class LineChars(Enum):

    ASCII = {"top_left": "+",
             "top_right": "+",
             "bottom_left": "+",
             "bottom_right": "+",
             "center_cross": "+",
             "left_cross": "+",
             "right_cross": "+",
             "top_cross": "+",
             "bottom_cross": "+",
             "horizontal": "-",
             "vertical": "|",
             "diagonal_fwd": "/",
             "diagonal_bwd": "\\",
             "diagonal_cross": "X"}

    BLOCK = {"top_left": "\u2588",
             "top_right": "\u2588",
             "bottom_left": "\u2588",
             "bottom_right": "\u2588",
             "center_cross": "\u2588",
             "left_cross": "\u2588",
             "right_cross": "\u2588",
             "top_cross": "\u2588",
             "bottom_cross": "\u2588",
             "horizontal": "\u2588",
             "vertical": "\u2588",
             "diagonal_fwd": "\u2588",
             "diagonal_bwd": "\u2588",
             "diagonal_cross": "\u2588"}

    SINGLE = {"top_left": "\u250C",
              "top_right": "\u2510",
              "bottom_left": "\u2514",
              "bottom_right": "\u2518",
              "center_cross": "\u253C",
              "left_cross": "\u251C",
              "right_cross": "\u2524",
              "top_cross": "\u252C",
              "bottom_cross": "\u2534",
              "horizontal": "\u2500",
              "vertical": "\u2502",
              "diagonal_fwd": "\u2571",
              "diagonal_bwd": "\u2572",
              "diagonal_cross": "\u2573"}

    DOUBLE = {"top_left": "\u2554",
              "top_right": "\u2557",
              "bottom_left": "\u255A",
              "bottom_right": "\u255D",
              "center_cross": "\u256C",
              "left_cross": "\u2560",
              "right_cross": "\u2563",
              "top_cross": "\u2566",
              "bottom_cross": "\u2569",
              "horizontal": "\u2550",
              "vertical": "\u2551",
              "diagonal_fwd": "\u2571",
              "diagonal_bwd": "\u2572",
              "diagonal_cross": "\u2573"}

    def __init__(self, char_set: Dict[str, str]):
        self._char_set = char_set

    @property
    def all_chars(self) -> Dict[str, str]:
        """Returns the full character dictionary."""
        return self._char_set

    @property
    def top_left(self) -> str:
        """Return the top left corner character."""
        return self._char_set["top_left"]

    @property
    def top_right(self) -> str:
        """Return the top right corner character."""
        return self._char_set["top_right"]

    @property
    def bottom_left(self) -> str:
        """Return the bottom left corner character."""
        return self._char_set["bottom_left"]

    @property
    def bottom_right(self) -> str:
        """Return the bottom right corner character."""
        return self._char_set["bottom_right"]

    @property
    def center_cross(self) -> str:
        """Return the intersection of horizontal and vertical character."""
        return self._char_set["center_cross"]

    @property
    def left_cross(self) -> str:
        """Return the intersection of the left side and horizontal character."""
        return self._char_set["left_cross"]

    @property
    def right_cross(self) -> str:
        """Return the intersection of the horizontal and right side character."""
        return self._char_set["right_cross"]

    @property
    def top_cross(self) -> str:
        """Return the intersection of the top side and vertical character."""
        return self._char_set["top_cross"]

    @property
    def bottom_cross(self) -> str:
        """Return the intersection of the vertical and bottom side character."""
        return self._char_set["bottom_cross"]

    @property
    def horizontal(self) -> str:
        """Return the horizontal character."""
        return self._char_set["horizontal"]

    @property
    def vertical(self) -> str:
        """Return the vertical character."""
        return self._char_set["vertical"]

    @property
    def diagonal_fwd(self) -> str:
        """Return a 'forward slash' (/) character."""
        return self._char_set["diagonal_fwd"]

    @property
    def diagonal_bwd(self) -> str:
        """Return a 'backward slash' (\\) character."""
        return self._char_set["diagonal_bwd"]

    @property
    def diagonal_cross(self) -> str:
        """Return the intersection of a 'forward slash' and 'backward slash' (X) character."""
        return self._char_set["diagonal_cross"]



from enum import Enum
from easyansi.drawing.line_chars import LineChars
from easyansi.common import field_validations as _validator
from easyansi.core import cursor as cur

MIN_LENGTH = 1


class Lines(Enum):

    ASCII = LineChars.ASCII
    BLOCK = LineChars.BLOCK
    SINGLE = LineChars.SINGLE
    DOUBLE = LineChars.DOUBLE

    def __init__(self, char_set: LineChars):
        self._char_set = char_set

    def horizontal(self, length: int) -> str:
        """Return a line of horizontal characters.

        Parameters:
            length: The length of the line to draw.
        """
        _validator.check_int_minimum_value(length, MIN_LENGTH, "Horizontal Line Length")
        return self._char_set.horizontal * length

    def vertical(self, length: int) -> str:
        """Return a line of vertical characters.

        Parameters:
            length: The length of the line to draw.
        """
        _validator.check_int_minimum_value(length, MIN_LENGTH, "Vertical Line Length")
        return (self._char_set.vertical + cur.move(-1, 1)) * (length - 1) + self._char_set.vertical



from easyansi.core import core as _core

# Font "Brightness"
# Bright and Dim send a normal ANSI code first because some terminals require this.
NORMAL = f"{_core.CSI}22m"

BRIGHT = f"{NORMAL}{_core.CSI}1m"
BRIGHT_ON = BRIGHT
BRIGHT_OFF = NORMAL

DIM = f"{NORMAL}{_core.CSI}2m"
DIM_ON = DIM
DIM_OFF = NORMAL

ITALIC = f"{_core.CSI}3m"
ITALIC_ON = ITALIC
ITALIC_OFF = f"{_core.CSI}23m"

UNDERLINE = f"{_core.CSI}4m"
UNDERLINE_ON = UNDERLINE
UNDERLINE_OFF = f"{_core.CSI}24m"

BLINK = f"{_core.CSI}5m"
BLINK_ON = BLINK
BLINK_OFF = f"{_core.CSI}25m"

REVERSE = f"{_core.CSI}7m"
REVERSE_ON = REVERSE
REVERSE_OFF = f"{_core.CSI}27m"

CONCEAL = f"{_core.CSI}8m"
CONCEAL_ON = CONCEAL
CONCEAL_OFF = f"{_core.CSI}28m"

STRIKETHROUGH = f"{_core.CSI}9m"
STRIKETHROUGH_ON = STRIKETHROUGH
STRIKETHROUGH_OFF = f"{_core.CSI}29m"

# This turns all above defined attributes off
ATTRIBUTES_OFF = f"{NORMAL}{ITALIC_OFF}{UNDERLINE_OFF}{BLINK_OFF}{REVERSE_OFF}{CONCEAL_OFF}{STRIKETHROUGH_OFF}"



# from collections.abc import Iterable
from typing import Any, Union, List, Tuple
import re


def _raise_error(field: str, value: str, msg: str, error_type: Any) -> None:
    """Build an error message, and use it to raise the error_type.

    Parameters:
        field: The name of the field in error.
        value: The value of the field in error.
        msg: The error message to show.
        error_type: The type of error / exception to raise.
    """
    raise error_type(f"Field: {field}, Value: {value}, Error: {msg}")


def check_if_boolean(value: bool, field: str) -> None:
    """Check if a value is a boolean.

    Parameters:
        value: The value to check for being a boolean.
        field: The name of the field being checked.
    """
    if not isinstance(value, bool):
        _raise_error(field, str(value), "Not a boolean value", TypeError)


def check_if_string(value: str, field: str) -> None:
    """Check if a value is a string.

    Parameters:
        value: The value to check for being a string.
        field: The name of the field being checked.
    """
    if not isinstance(value, str):
        _raise_error(field, str(value), "Not a string value", TypeError)


def check_if_int(value: int, field: str) -> None:
    """Check if a value is an int.

    Parameters:
        value: The value to check for being an int.
        field: The name of the field being checked.
    """
    if (not isinstance(value, int)) or (isinstance(value, bool)):
        _raise_error(field, str(value), "Not an int value", TypeError)


def check_int_minimum_value(value: int, minimum: int, field: str) -> None:
    """Check if a value is >= a minimal value.

    Parameters:
        value: The value to check.
        minimum: The minimum the value can be.
        field: The name of the field being checked.
    """
    check_if_int(value, field)
    check_if_int(minimum, f"Minimum for {field}")
    if not value >= minimum:
        _raise_error(field, str(value), f"Value is less than minimum of {minimum}", ValueError)


def check_int_in_range(value: int, minimum: int, maximum: int, field: str) -> None:
    """Check if an integer is between 2 values.

    Parameters:
        value: The value to check for being within the range.
        minimum: The minimum value the value can be.
        maximum: The maximum value the value can be.
        field: The name of the field being checked.
    """
    check_if_int(value, field)
    check_if_int(minimum, f"Minimum for {field}")
    check_if_int(maximum, f"Maximum for {field}")
    if not minimum <= value <= maximum:
        _raise_error(field, str(value), f"Value is not within range of {minimum} to {maximum}", ValueError)


def check_if_any_have_value(field: str, *args: Any, zero_int_no_value: bool = False) -> None:
    """Given an arbitrary number of arguments, raise an error if none of them have a value.

    Parameters:
        field: The name of the field(s) being checked.
        *args: values to be checked to be anything except None.
        zero_int_no_value: True/False: if an integer is zero, count this as no value.
    """
    value_found = False
    for arg in args:
        if arg is not None:
            if zero_int_no_value:
                if (isinstance(arg, int)) or (isinstance(arg, bool)):
                    if arg != 0:
                        value_found = True
                else:
                    value_found = True
            else:
                value_found = True
        if value_found:
            break
    if not value_found:
        _raise_error(field, "", "No provided arguments have a value", ValueError)


def check_if_string_matches_regex(value: str, regex: str, field: str) -> None:
    """Given a string and a regex, check if the string matches the regex.

    Parameters:
        value: The string value to check.
        regex: The regular expression to test the string against.
        field: The name of the field being checked.
    """
    check_if_string(value, field)
    check_if_string(regex, f"Regex for {field}")
    if not re.match(regex, value):
        _raise_error(field, str(value), "Value is not in a valid format", ValueError)


def check_if_sequence(value: Union[List[Any], Tuple[Any]], field: str) -> None:
    """Given a value, check if it is a supported sequence.

    Parameters:
        value: The value to check to be a supported sequence.
        field: The name of the field being checked.
    """
    if not ((isinstance(value, list)) or (isinstance(value, tuple))):
        _raise_error(field, str(value), "Value is not a supported sequence", TypeError)



from easyansi.core.color_builder import ColorBuilder as _ColorBuilder
_builder = _ColorBuilder.PALETTE_16._get_color_code_by_int

DEFAULT_IDX = -1
DEFAULT_FG = _builder(DEFAULT_IDX, False)
DEFAULT_BG = _builder(DEFAULT_IDX, True)
DEFAULT = DEFAULT_FG

BLACK_IDX = 0
BLACK_FG = _builder(BLACK_IDX, False)
BLACK_BG = _builder(BLACK_IDX, True)
BLACK = BLACK_FG

RED_IDX = 1
RED_FG = _builder(RED_IDX, False)
RED_BG = _builder(RED_IDX, True)
RED = RED_FG

GREEN_IDX = 2
GREEN_FG = _builder(GREEN_IDX, False)
GREEN_BG = _builder(GREEN_IDX, True)
GREEN = GREEN_FG

YELLOW_IDX = 3
YELLOW_FG = _builder(YELLOW_IDX, False)
YELLOW_BG = _builder(YELLOW_IDX, True)
YELLOW = YELLOW_FG

BLUE_IDX = 4
BLUE_FG = _builder(BLUE_IDX, False)
BLUE_BG = _builder(BLUE_IDX, True)
BLUE = BLUE_FG

MAGENTA_IDX = 5
MAGENTA_FG = _builder(MAGENTA_IDX, False)
MAGENTA_BG = _builder(MAGENTA_IDX, True)
MAGENTA = MAGENTA_FG

CYAN_IDX = 6
CYAN_FG = _builder(CYAN_IDX, False)
CYAN_BG = _builder(CYAN_IDX, True)
CYAN = CYAN_FG

WHITE_IDX = 7
WHITE_FG = _builder(WHITE_IDX, False)
WHITE_BG = _builder(WHITE_IDX, True)
WHITE = WHITE_FG

BRIGHT_BLACK_IDX = 8
BRIGHT_BLACK_FG = _builder(BRIGHT_BLACK_IDX, False)
BRIGHT_BLACK_BG = _builder(BRIGHT_BLACK_IDX, True)
BRIGHT_BLACK = BRIGHT_BLACK_FG

BRIGHT_RED_IDX = 9
BRIGHT_RED_FG = _builder(BRIGHT_RED_IDX, False)
BRIGHT_RED_BG = _builder(BRIGHT_RED_IDX, True)
BRIGHT_RED = BRIGHT_RED_FG

BRIGHT_GREEN_IDX = 10
BRIGHT_GREEN_FG = _builder(BRIGHT_GREEN_IDX, False)
BRIGHT_GREEN_BG = _builder(BRIGHT_GREEN_IDX, True)
BRIGHT_GREEN = BRIGHT_GREEN_FG

BRIGHT_YELLOW_IDX = 11
BRIGHT_YELLOW_FG = _builder(BRIGHT_YELLOW_IDX, False)
BRIGHT_YELLOW_BG = _builder(BRIGHT_YELLOW_IDX, True)
BRIGHT_YELLOW = BRIGHT_YELLOW_FG

BRIGHT_BLUE_IDX = 12
BRIGHT_BLUE_FG = _builder(BRIGHT_BLUE_IDX, False)
BRIGHT_BLUE_BG = _builder(BRIGHT_BLUE_IDX, True)
BRIGHT_BLUE = BRIGHT_BLUE_FG

BRIGHT_MAGENTA_IDX = 13
BRIGHT_MAGENTA_FG = _builder(BRIGHT_MAGENTA_IDX, False)
BRIGHT_MAGENTA_BG = _builder(BRIGHT_MAGENTA_IDX, True)
BRIGHT_MAGENTA = BRIGHT_MAGENTA_FG

BRIGHT_CYAN_IDX = 14
BRIGHT_CYAN_FG = _builder(BRIGHT_CYAN_IDX, False)
BRIGHT_CYAN_BG = _builder(BRIGHT_CYAN_IDX, True)
BRIGHT_CYAN = BRIGHT_CYAN_FG

BRIGHT_WHITE_IDX = 15
BRIGHT_WHITE_FG = _builder(BRIGHT_WHITE_IDX, False)
BRIGHT_WHITE_BG = _builder(BRIGHT_WHITE_IDX, True)
BRIGHT_WHITE = BRIGHT_WHITE_FG



import re
from easyansi.common import field_validations as _validator


def prnt(text: str) -> None:
    """This is a wrapper around print which forces a buffer flush and does not output a newline.

    Parameters:
        text: The text to print."""
    print(text, flush=True, end='')


def prntln(text: str = '') -> None:
    """This is a wrapper around print which forces a buffer flush and does output a newline.

    Parameters:
        text: The text to print."""
    print(text, flush=True)


def strip_ansi_codes(text: str) -> str:
    """Given a text string, remove all ANSI sequences and return the plain text.

    This is only guaranteed to remove ANSI codes generated by Easy ANSI, though the following should also work:
    * Any ANSI CSI sequence.

    Parameters:
        text: The text to strip ANSI codes out of."""

    # Source of CSI format: https://en.wikipedia.org/wiki/ANSI_escape_code
    #   CSI Starting Sequence = Esc + [
    #   CSI Parameter Bytes: 0–9 : ; < = > ?
    #   CSI Intermediate Bytes: [space] ! " # $ % & ' ( ) * + , - . /
    #   CSI Final Byte: A-Z a-z @ [ \ ] ^ _ ` { | } ~

    _validator.check_if_string(text, "Text to strip ANSI codes from")
    csi_regex = r'(\x1B\[)[0-?]*[ -\/]*[@-~]'
    csi_pattern = re.compile(csi_regex)  # compile csi regex pattern
    csi_less_text = csi_pattern.sub('', text)  # replace regex matches with empty string
    return csi_less_text


def text_len(text: str) -> int:
    """Given a text string, return the length of the text without the ANSI characters.

    Parameters:
        text: The text to calculate the length from."""
    return len(strip_ansi_codes(text))



from setuptools import setup, find_packages
from pathlib import Path

version = '2.0.0'

readme_file = '../README.md'

description = ("An ANSI API framework with no dependencies on 3rd-party libraries, with an emphasis on not needing " +
               "to understand the details of ANSI escape sequences themselves.")

long_description = Path(readme_file).read_text()

# https://pypi.org/classifiers/
classifiers = [
    'Development Status :: 3 - Alpha',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Topic :: Software Development',
    'Topic :: Software Development :: Libraries',
    'Topic :: Software Development :: User Interfaces',
    'Topic :: Terminals',
    'Typing :: Typed'
]

keywords = 'ANSI, easy, API, console, terminal'

project_urls = {
    'Source': 'https://gitlab.com/easy-ansi/easy-ansi',
    'Documentation': 'https://gitlab.com/easy-ansi/easy-ansi-docs',
    'Issues': 'https://gitlab.com/easy-ansi/easy-ansi/-/issues',
    'Demos': 'https://gitlab.com/easy-ansi/easy-ansi-demos'
}

packages = find_packages(include=['easyansi', 'easyansi.*'])

setup(
    name='easy-ansi',
    version=version,
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/easy-ansi/easy-ansi',
    author='Joey Rockhold',
    author_email='joey@joeysbytes.net',
    classifiers=classifiers,
    keywords=keywords,
    packages=packages,
    project_urls=project_urls,
    python_requires='>=3.6,<4'
)
